#define LEAF 0

typedef struct NodeRel {
    int id; 
    // Node-Id. ids entsprechen Position im Array
    // (Diese hier nicht unbedingt ntig)
    
    int lChild; // id der linken Child-Node
    
    int rChild; // id der rechten Child Node
    
    int circleMapOffset; // offset zum Zugriff in mapNode2Circles
    // ntig, um zugehrige Circles zu finden
} NodeRel;


typedef struct Circle {
    float x;
    float y;
    float r;
    float depth;
    float4 color;
} Circle;

typedef struct BBox2D {
    float minX; // minimaler und
    float maxX; // maximaler x-Wert
    float minY; // minimaler und
    float maxY; // maximaler y-Wert
} BBox2D;


NodeRel getContainingChild(BBox2D bBoxLeft, // Bbox des linken Childs
NodeRel cL, // Linke Child Node
NodeRel cR, // Rechte Child Node
int2 pixelPos // Position des Pixels
){
    if( bBoxLeft.minX <= pixelPos.x // Teste,
        && bBoxLeft.maxX > pixelPos.x // ob im
        && bBoxLeft.minY <= pixelPos.y // linken Child
        && bBoxLeft.maxY > pixelPos.y // enthalten
        )
        return cL; // Ja, liefere dieses zurck
    else
        return cR; // Sonst muss es im anderen enthalten sein
}

bool overlaps_N_Nearer(Circle c, int2 pixelPos, int minDepth){
    // Werte implizite Kreisgleichung fr Position des Pixels aus
    float dx = (float) c.x - pixelPos.x;
    float dy = (float) c.y - pixelPos.y;
    return (dx * dx + dy * dy < c.r * c.r) && (minDepth > c.depth);
}

__kernel void implicit2DRenderer_KD(
__write_only image2d_t rederedImage,
__global NodeRel* nodes_Relations,
__global int* nodes_flag,
__global BBox2D* nodes_BBox,
__global Circle* circles,
__global int* mapNode2Circles) {
    
    int2 pPos = (int2)(get_global_id(0), get_global_id(1));
    NodeRel node = nodes_Relations[0]; // Beginne mit Root Node
    while(nodes_flag[node.id] != LEAF) { // Traversiere KD-Tree bis LEAF gefunden
        NodeRel cL = nodes_Relations[node.lChild];
        NodeRel cR = nodes_Relations[node.rChild];
        node = getContainingChild(nodes_BBox[cL.id], cL, cR, pPos); // Folie 12
    }
    float minDepth = 100000; float4 color = (float4) (0.0, 0.0, 0.0, 0.0); // Hintergrund
    int nodeCircleCnt = mapNode2Circles[node.circleMapOffset];
    for(int i = 1; i <= nodeCircleCnt ; ++i) {
        Circle c = circles[mapNode2Circles[node.circleMapOffset + i]];
        if(overlaps_N_Nearer(c, pPos, minDepth)) { // Folie 29
            minDepth = c.depth;
            color = c.color;
        }
    }
    write_imagef(rederedImage, pPos, color);
}