const sampler_t samplerA = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_REPEAT | CLK_FILTER_LINEAR;
#define LOCAL_MEMORY 16 * 16 //size*size
typedef struct Circle {
    float4 posr;
    float4 colord;
} Circle;

__kernel void ray2dShared(
__global float4* texture,
__global Circle* data,
int count) 
{

    uint xid = get_global_id(0);
    uint yid = get_global_id(1);
    uint localx = get_local_id(0);
    uint localy = get_local_id(1);
    uint local_work_size_x = get_local_size(0);
    uint local_work_size_y = get_local_size(1);
    uint localPos = localy*local_work_size_x + localx;
    uint bufferPos = yid * get_global_size(0) + xid;
    float minDepth = -1;
    __local Circle localData[LOCAL_MEMORY];
    uint tile = local_work_size_x*local_work_size_y;
    for(uint i = 0; i < count; i+= tile)
    {
        localData[localPos] = data[i + localPos];
        barrier(CLK_LOCAL_MEM_FENCE);
        for(uint j = 0; j < tile; ++j)
        {
            Circle c = localData[j];
            float x = c.posr.x - xid;
            float y = c.posr.y - yid;
            if(x*x + y*y < c.posr.w*c.posr.w) {
               if(minDepth > c.colord.w || minDepth == -1) {
                    minDepth = c.colord.w;
                    float4 color = (float4)(c.colord.x, c.colord.y, c.colord.z, 0);
                    texture[bufferPos] = color;
               }
            }
        }
        barrier(CLK_LOCAL_MEM_FENCE);
    }
}

__kernel void ray2d(
__global float4* texture,
__global Circle* data,
int count) 
{
    uint xid = get_global_id(0);
    uint yid = get_global_id(1);
    uint w = get_global_size(0);
    uint h = get_global_size(1);
    uint pos = yid * w + xid;
    float minDepth = -1;
    for(uint i = 0; i < count; ++i)
    {
        Circle c = data[i];
        float x = c.posr.x - xid;
        float y = c.posr.y - yid;
        if(x*x + y*y < c.posr.w*c.posr.w) {
            if(minDepth > c.colord.w || minDepth == -1) {
                minDepth = c.colord.w;
                float4 color = (float4)(c.colord.x, c.colord.y, c.colord.z, 0);
                texture[pos] = color;
           }
        }
    }
}