package kd;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.lwjgl.BufferUtils;
import org.lwjgl.LWJGLException;
import org.lwjgl.PointerBuffer;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CL10GL;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLDevice;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLPlatform;
import org.lwjgl.opencl.CLProgram;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.util.vector.Vector4f;

import util.OpenCL;
import util.Util;


public class RayTracer2DKD {

	private static int w = 512, h = 512;
	private static boolean done = false;
	private static PointerBuffer gws = BufferUtils.createPointerBuffer(2);
	private static PointerBuffer lws = BufferUtils.createPointerBuffer(2);
	private static int circleCount = 4096;
	private static int maxTreeDepth = 12;
	private static int targetElementsPerNode = 1;
    
	private static CLContext context;
	private static CLDevice device;
	private static CLProgram program;
	private static CLCommandQueue queue;
       
    private static CLMem nodesMem, circlesMem, nodesContentMem, flagsMem, texMem, minMaxMem;   
    private static CLKernel kernel;
    
    private static final int NODE_SIZE  = 4;
    private static final int CIRCLE_SIZE = 8;
    
    private static int texture;
    
    private static final int LEFT = 1;
    private static final int RIGHT = 2;
    private static final int LEAF = 0;
    
    private static final List<Node> nodes = new ArrayList<>();
    
    private static final List<Circle> circles = new ArrayList<>();
    
    static class Circle {
        Vector4f posr = new Vector4f();
        Vector4f colord = new Vector4f();
        Vector4f minMax = new Vector4f();
        
        public int getId() {
            return circles.indexOf(this);
        }
    }
    
    static class Node {
        Vector4f minMax = new Vector4f();
        List<Circle> circles = new ArrayList<>();
        Node parent;
        Node c0;
        Node c1;
        int axis = -1;
        int flag = -1;
        int contentOffSet;
        public String toString() {
            return "minMax="+minMax +
                    ", id=" + this.getId() + 
            		", parent="+((parent == null) ? -1 : nodes.indexOf(parent)) + 
            		", flag=" + flag + 
            		", count=" + circles.size() +
            		", contentOffset=" + contentOffSet;
        }
        
        public int getId() {
            return nodes.indexOf(this);
        }
    }

    
    static void createCL(String source) throws LWJGLException {
        
        CLPlatform platform = null;
        
        for(CLPlatform plf : CLPlatform.getPlatforms()) {
            if(plf.getDevices(CL10.CL_DEVICE_TYPE_GPU) != null) {
                device = plf.getDevices(CL10.CL_DEVICE_TYPE_GPU).get(0);
                platform = plf;
                if(device != null) {
                    break;
                }
            }
        }
        
        context = CLContext.create(platform, platform.getDevices(CL10.CL_DEVICE_TYPE_GPU), null, Display.getDrawable(), null);
        
        queue = OpenCL.clCreateCommandQueue(context, device, 0);
        
        program = OpenCL.clCreateProgramWithSource(context, source);
        OpenCL.clBuildProgram(program, device, "", null);
        OpenCL.checkProgram(program, device);
    }
    
    public static void main(String[] args) {
        try {
            Display.setDisplayMode(new DisplayMode(w, h));
            Display.create();
            Keyboard.create();
            Mouse.create();
            OpenCL.init();
            Util.init();
            createCL(Util.getText("./data/simpleKDRayT.cl"));
        } catch (LWJGLException e) {
            e.printStackTrace();
        }
        
        texture = GL11.glGenTextures();
        GL13.glActiveTexture(GL13.GL_TEXTURE0 + 1);
        GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture);
        GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, GL11.GL_RGBA8, w, h, 0, GL11.GL_RGBA, GL11.GL_FLOAT, (FloatBuffer)null);
        GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
        GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        GL13.glActiveTexture(GL13.GL_TEXTURE0);

        Random r = new Random(1);
        FloatBuffer cb = BufferUtils.createFloatBuffer(circleCount * CIRCLE_SIZE);
        for(int i = 0; i < circleCount; ++i) {
            Circle c = new Circle();
            c.posr.x =  r.nextInt(w); //x
            c.posr.y =  r.nextInt(w); //y
            c.posr.z =  5; //radius
            c.posr.w =  r.nextFloat(); //r
            c.colord.x =  r.nextFloat(); //g
            c.colord.y =  r.nextFloat(); //b
            c.colord.z =  r.nextFloat(); //a
            c.colord.w =  r.nextFloat(); //d
            c.minMax = getMinMax(c);
            circles.add(c);
            
            cb.put(c.posr.x);
            cb.put(c.posr.y);
            cb.put(c.posr.z);
            cb.put(c.posr.w);
            cb.put(c.colord.x);
            cb.put(c.colord.y);
            cb.put(c.colord.z);
            cb.put(c.colord.w);
        }    
        cb.position(0);
        
        createTree(null, 0);
        
        int contentSize = 0;
        List<Integer> contentl = new ArrayList<>();
        for(Node n : nodes) {
            if(n.flag == LEAF) {
                contentl.add(n.circles.size());
                for(Circle c : n.circles) {
                    contentl.add(c.getId()); 
                }
                n.contentOffSet = contentSize;
                contentSize += n.circles.size() + 1;
            }
        }
        
        int[] contentArray = new int[contentSize];
        for(int i = 0; i < contentSize; ++i) {
            contentArray[i] = contentl.get(i);
        }
        
        int nodeArray[] = new int[nodes.size()*NODE_SIZE];

        int index = 0;
        for(int i = 0; i < nodes.size(); ++i) {
            Node n = nodes.get(i);
            nodeArray[index++] = n.getId();
            Node c0 = n.c0;
            nodeArray[index++] = c0 == null ? -1 : c0.getId();
            Node c1 = n.c1;
            nodeArray[index++] = c1 == null ? -1 : c1.getId();
            nodeArray[index++] = n.contentOffSet;
        }
        
        int[] flagArray = new int[nodes.size()];
        for(int i = 0; i < nodes.size(); ++i) {
            flagArray[i] = nodes.get(i).flag;
        }
        
        float mm[] = new float[nodes.size() * 4];
        for(int i = 0; i < nodes.size(); ++i) {
            Node n = nodes.get(i);
            mm[4*i+0] = n.minMax.x;
            mm[4*i+1] = n.minMax.y;
            mm[4*i+2] = n.minMax.z;
            mm[4*i+3] = n.minMax.w;
        }
        
        IntBuffer nodes = BufferUtils.createIntBuffer(nodeArray.length);
        nodes.put(nodeArray);
        nodes.position(0);
        
        IntBuffer flags = BufferUtils.createIntBuffer(flagArray.length);
        flags.put(flagArray);
        flags.position(0);
        
        IntBuffer content = BufferUtils.createIntBuffer(contentArray.length);
        content.put(contentArray);
        content.position(0);
        
        FloatBuffer minMax = BufferUtils.createFloatBuffer(mm.length);
        minMax.put(mm);
        minMax.position(0);
        
        nodesMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, nodes);
        
        circlesMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, cb);
        
        nodesContentMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, content);
        
        flagsMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, flags);
        
        minMaxMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, minMax);
        
        texMem = CL10GL.clCreateFromGLTexture2D(context, CL10.CL_MEM_WRITE_ONLY, GL11.GL_TEXTURE_2D, 0, texture, null);
                        
        kernel = OpenCL.clCreateKernel(program, "implicit2DRenderer_KD");
        
        kernel.setArg(0, texMem);
        kernel.setArg(1, nodesMem);
        kernel.setArg(2, flagsMem);
        kernel.setArg(3, minMaxMem);
        kernel.setArg(4, circlesMem);
        kernel.setArg(5, nodesContentMem);

        gameLoop();
    }
    
    static int nodeCount = 0;
    static FloatBuffer b = BufferUtils.createFloatBuffer(1);
    
    static int getLongestAxis(float w, float h) {
        return w < h ? 1 : 0;
    }
    
    static void createTree(Node parent, int depth) {
        if(parent == null) {
            parent = new Node();
            parent.minMax.y = w;
            parent.minMax.w = h;
            parent.circles = circles;
            parent.flag = -1;
            nodes.add(parent);
        }
        
        if(depth == maxTreeDepth) {
            parent.flag = LEAF;
            return;
        }
        
        Node c0 = new Node();
        c0.parent = parent;
        Node c1 = new Node();
        c1.parent = parent;
        parent.c0 = c0;
        parent.c1 = c1;
        nodes.add(c0);
        nodes.add(c1);

        float c0minx = parent.minMax.x;
        float c0maxx = parent.minMax.y;
        float c0miny = parent.minMax.z;
        float c0maxy = parent.minMax.w;
        
        float c1minx = c0minx;
        float c1maxx = c0maxx;
        float c1miny = c0miny;
        float c1maxy = c0maxy;
        
        int axis = getLongestAxis(c0maxx - c0minx, c0maxy - c0miny);
        switch(axis) {
        case 0: {
            float cutX = c0minx + (c0maxx - c0minx) * 0.5f;
            c0maxx = cutX;
            c1minx = cutX;
            
        } break;
        case 1: {
            float cutY = c0miny + (c0maxy - c0miny) * 0.5f;
            c0maxy = cutY;
            c1miny = cutY;
            
        } break;
        }
        
        assert c0minx != c0maxx;
        assert c0miny != c0maxy;
        assert c1minx != c1maxx;
        assert c1miny != c1maxy;

        c0.minMax.x = c0minx;
        c0.minMax.y = c0maxx;
        c0.minMax.z = c0miny;
        c0.minMax.w = c0maxy;
        c0.parent = parent;
        c0.axis = axis;
        c0.flag = LEFT;
        
        c1.minMax.x = c1minx;
        c1.minMax.y = c1maxx;
        c1.minMax.z = c1miny;
        c1.minMax.w = c1maxy;
        c1.parent = parent;
        c1.axis = axis;
        c1.flag = RIGHT;
               
        arrangeCirlces(parent, c0);
        arrangeCirlces(parent, c1);        
       
        if(c0.circles.size() > targetElementsPerNode) {
            createTree(c0, depth+1); 
        } else {
            c0.flag = LEAF;
        }

        if(c1.circles.size() > targetElementsPerNode) {
            createTree(c1, depth+1); 
        } else {
            c1.flag = LEAF;
        }
    }
    
    static void arrangeCirlces(Node parent, Node child) {
        int circleCount = parent.circles.size();
        int flag = child.flag;
        for(int i = 0; i < circleCount; ++i) {
            int axis = child.axis;
            Circle c = parent.circles.get(i);
            if(flag == RIGHT) {
                if(getMaxAxis(axis, c.minMax) > getMinAxis(axis, child.minMax)) {
                    child.circles.add(c);
                }
            } else if(flag == LEFT){
                if(getMinAxis(axis, c.minMax) < getMaxAxis(axis, child.minMax)) {
                    child.circles.add(c);
                }
            }
        }
    }
    
    static Vector4f getMinMax(Circle circle) {
        Vector4f minMax = new Vector4f();
        minMax.x = circle.posr.x - circle.posr.z;
        minMax.y = circle.posr.x + circle.posr.z;
        
        minMax.z = circle.posr.y - circle.posr.z;
        minMax.w = circle.posr.y + circle.posr.z;
        return minMax;
    }
    
    static float getMinAxis(int axis, Vector4f minMax) {
        switch(axis) {
        case 0 : return minMax.x;
        case 1 : return minMax.z;
        }
        throw new IllegalArgumentException();
    }
    
    static float getMaxAxis(int axis, Vector4f minMax) {
        switch(axis) {
        case 0 : return minMax.y;
        case 1 : return minMax.w;
        }
        throw new IllegalArgumentException();
    }
 
    static void gameLoop() {       
        gws.put(0, w);
        gws.put(1, h);
        lws.put(0, 8);
        lws.put(1, 8);
        while(!done && !Display.isCloseRequested()) {
            if(Keyboard.isKeyDown(Keyboard.KEY_ESCAPE)) {
                done = true;
            }
            Display.update();

            OpenCL.clEnqueueAcquireGLObjects(queue, texMem, null, null);
            OpenCL.clEnqueueNDRangeKernel(queue, kernel, 2, null, gws, lws, null, null);
            OpenCL.clEnqueueReleaseGLObjects(queue, texMem,  null, null);
            OpenCL.clFinish(queue);
            GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
            Util.drawTexture(1);
            
            Util.tick();
            Util.printFps();
        }
        clean();
    }
    
    static void clean()  {
        OpenCL.clReleaseMemObject(nodesMem);
        OpenCL.clReleaseMemObject(circlesMem);
        OpenCL.clReleaseMemObject(nodesContentMem);
        OpenCL.clReleaseMemObject(flagsMem);
        OpenCL.clReleaseMemObject(minMaxMem);
        OpenCL.clReleaseMemObject(texMem);
        OpenCL.clReleaseKernel(kernel);
        OpenCL.clReleaseCommandQueue(queue);
        OpenCL.clReleaseProgram(program);
        OpenCL.clReleaseContext(context);
        OpenCL.destroy();
        Mouse.destroy();
        Keyboard.destroy();
        Util.destroy();
        Display.destroy();
    }
}
