package nokd;
import java.nio.FloatBuffer;
import java.util.Random;

import org.lwjgl.BufferUtils;
import org.lwjgl.LWJGLException;
import org.lwjgl.PointerBuffer;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLDevice;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLPlatform;
import org.lwjgl.opencl.CLProgram;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;

import util.OpenCL;
import util.Util;


public class RayTracer2D {
    static FloatBuffer circles;
    static int w = 512, h = 512;
    static boolean done = false;
    static PointerBuffer gws = BufferUtils.createPointerBuffer(2);
    static PointerBuffer lws = BufferUtils.createPointerBuffer(2);
    static boolean sharedMemory = true;
    static int count = 4096;
    
    static CLContext context;
    static CLDevice device;
    static CLProgram program;
    static CLKernel kernelShared, kernel;
    static CLCommandQueue queue;
    
    static int texture;
    static CLMem texMem, circleMem;
    static void createCL(String source) throws LWJGLException {
        
        CLPlatform platform = null;
        
        for(CLPlatform plf : CLPlatform.getPlatforms()) {
            if(plf.getDevices(CL10.CL_DEVICE_TYPE_GPU) != null) {
                device = plf.getDevices(CL10.CL_DEVICE_TYPE_GPU).get(0);
                platform = plf;
                if(device != null) {
                    break;
                }
            }
        }
        
        context = CLContext.create(platform, platform.getDevices(CL10.CL_DEVICE_TYPE_GPU), null, Display.getDrawable(), null);
        
        queue = OpenCL.clCreateCommandQueue(context, device, 0);
        
        program = OpenCL.clCreateProgramWithSource(context, source);
        OpenCL.clBuildProgram(program, device, "", null);
        OpenCL.checkProgram(program, device);
    }
    
    public static void main(String[] args) {
        
        try {
            Display.setDisplayMode(new DisplayMode(w, h));
            Display.create();
            Keyboard.create();
            Mouse.create();
            OpenCL.init();
            createCL(Util.getText("./data/simpleRayT.cl"));
        } catch (LWJGLException e) {
            e.printStackTrace();
        }
        texMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_WRITE_ONLY, w * h * 16);
        
        circles = BufferUtils.createFloatBuffer(count * 8);
        Random r = new Random(1);
        for(int i = 0; i < count; ++i) {
            setCircle(r.nextInt(w), r.nextInt(h), 5, r.nextFloat(), r.nextFloat(), r.nextFloat(), 1, r.nextFloat());
        }
        circles.flip();
        circleMem = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, circles);
        
        
        kernel = OpenCL.clCreateKernel(program, "ray2d");
        kernelShared = OpenCL.clCreateKernel(program, "ray2dShared");
        
        kernel.setArg(0, texMem);
        kernel.setArg(1, circleMem);
        kernel.setArg(2, count);
        
        kernelShared.setArg(0, texMem);
        kernelShared.setArg(1, circleMem);
        kernelShared.setArg(2, count);
        
        gws.put(0, w);
        gws.put(1, h);
        lws.put(0, 16);
        lws.put(1, 16); 
        OpenCL.clFinish(queue);
        System.out.println(GL11.glGetError() == 0 ? "no gl error" : "gl error");
        gameLoop();
    }

    static void setCircle(float x, float y, float radius, float r, float g, float b, float a, float d) {
        circles.put(x);
        circles.put(y);
        circles.put(0);
        circles.put(radius);
        circles.put(r);
        circles.put(g);
        circles.put(b);
        circles.put(d);
    }

    static void gameLoop() {
        while(!done && !Display.isCloseRequested()) {
            if(Keyboard.isKeyDown(Keyboard.KEY_ESCAPE)) {
                done = true;
            }
            Util.tick();
            GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
            if(sharedMemory) {
                CL10.clEnqueueNDRangeKernel(queue, kernelShared, 2, null, gws, lws, null, null);
            } else {
                OpenCL.clEnqueueNDRangeKernel(queue, kernel, 2, null, gws, lws, null, null); 
            }
            Util.drawCLBuffer(queue, texMem, w, h, 4);
            OpenCL.clFinish(queue);
            Display.update();
            Util.printFps(40);
        }
        clean();
    }
    
    static void clean()  {
        GL11.glDeleteTextures(texture);
        OpenCL.clReleaseMemObject(texMem);
        OpenCL.clReleaseMemObject(circleMem);
        OpenCL.clReleaseCommandQueue(queue);
        OpenCL.clReleaseKernel(kernel);
        OpenCL.clReleaseKernel(kernelShared);
        OpenCL.clReleaseProgram(program);
        OpenCL.clReleaseContext(context);
        OpenCL.destroy();
        Mouse.destroy();
        Keyboard.destroy();
        Display.destroy();
    }
}
